package mdp.rtdp;

import graphics.BlockDisplay;

import java.awt.Color;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;

public class Racetrack {

	// Limit trajectories to 1000 steps
	public final static int MIN_DISP_BELLMAN_UPDATES = 1000000000;

	public final static int MAX_TRAJ_LENGTH = 200;
	public final static double VALUE_MAX = 0;
	public final static double VALUE_MIN = -MAX_TRAJ_LENGTH;

	public final static boolean USE_UPPER_BOUND_COMP = true;

	public static final double BRTDP_TOLERANCE = 1e-6d; 
	public static final double VPI_CUTOFF = 0.0d;
	public static double BOUND_GAP_CUTOFF = 190d;
	public static double ALPHA = 1e-3d;
	
	public static final boolean DRAW_RANDOM_VPI  = true;
		
	public static boolean SKIP_MARKED = true;
	public static boolean USE_INFORMED_HEURISTIC = true;

	public static boolean MAINTAIN_FRTDP_PRIORITY = false; // Only for FRTDP
	public static boolean MAINTAIN_QVALUES = false; // Only for VPI-RTDP

	public static DecimalFormat _df = new DecimalFormat("#.##");

	// Bookkeeping
	public Random _random = new Random();
	public String _filename = null;
	public BlockDisplay bd = null;
	public Track _track = null;
	public State _tempState = new State(0, 0, 0, 0);
	public TreeMap _states = new TreeMap();
	public ArrayList _startStates = new ArrayList();
	public Solver _compare = null;
	public PrintWriter _ps = null;
	public Pos _finalPos = null;

	// Used to keep track of unique next states and probabilities
	public class NextState {
		public NextState(double p, State s) {
			prob = p;
			state = s;
		}
		double prob;
		State  state;
		public int hashCode() {
			return state.hashCode();
		}
		public boolean equals(Object o) {
			return state.equals(((NextState)o).state);
		}
		public void add(NextState s) {
			if (!s.state.equals(this.state)) {
				System.out.println("Cannot add non-identical states!");
				System.exit(1);
			}
			this.prob += s.prob;
		}
	}

	// Returned by Bellman Update
	public static class BestAction {
		public BestAction(int lb /*double lb_v*/,  
				int avg /*, double avg_v,*/,  int ub /*, double ub_v*/) {
			lb_action = lb;
			//lb_val = lb_v;
			avg_action = avg;
			//avg_val = avg_v;
			ub_action = ub;
			//ub_val = ub_v;
		}
		int lb_action;
		//double lb_val;
		int avg_action;
		//double avg_val;
		int ub_action;
		//double ub_val;
	}

	////////////////////////////////////////////////////////////////////////////
	// Racetrack Representation
	////////////////////////////////////////////////////////////////////////////
	
	// The racetrack representation is just an array of Strings
	public class Track {
		String[] _layout = null;
		int _height = -1;
		int _width = -1;

		public Track(int height, int width) {
			_height = height;
			_width = width;
			_layout = new String[height];
		}

		public char get(int x, int y) {
			if (x < 0 || x >= _width || y < 0 || y >= _height)
				return '@';
			return _track._layout[y].charAt(x);
		}
	}

	////////////////////////////////////////////////////////////////////////////
	// State and Value Representation, Maintenance
	////////////////////////////////////////////////////////////////////////////

	public static String getActionString(int act) {
		int ddx = (act % 3) - 1;
		int ddy = ((act / 3) % 3) - 1;
		return "(" + ddx + ", " + ddy + ")";
	}
	
	static int _nCurTime = 0;

	public void clearMarks() {
		_nCurTime++;
	}
	
	static int _nBellmanUpdates = 0;
	
	// TODO: This State should be abstracted and some methods removed to 
	//       BRTDP / FRTDP / VPI-RTDP
	public class State implements Comparable {
		int x, y;
		int dx, dy;
		char ctype;
		double _valueUpper; // RMax
		double _valueLower; // RMin
		double _priority; // For FRTDP
		double _maxVPI = 0d;
		int _nMark;
		
		// First index is number of actions, second index is outcome index
		//
		// Actions: <ddx{-1,0,1},ddy{-1,0,1}>
		// num = ddx + 1 + (3 * (ddy + 1)) \in {0,...,8}
		// ddx = (num % 3) - 1 \in {-1,0,1}
		// ddy = ((num / 3) % 3) - 1 \in {-1,0,1}
		public State[/* act */][/* out */] _nextState = null;
		public double[/* act */][/* out */] _nextProb = null;
		public double[/* act */] _qUp  = null;
		public double[/* act */] _qLow = null;
		public double[/* act */] _qAvg = null;

		public State(int x, int y, int dx, int dy) {
			this.x = x;
			this.y = y;
			this.dx = dx;
			this.dy = dy;
			this.ctype = _track == null ? '@' : _track.get(x, y);
			if (this.ctype == 'f') {
				this._valueUpper = 0d;
				this._valueLower = 0d;
			} else {
				this._valueLower = VALUE_MIN;
				
				if (USE_INFORMED_HEURISTIC && _finalPos != null) {
					this._valueUpper = (Math.abs(_finalPos.x - x) + 
					                    Math.abs(_finalPos.y - y)) / 
					                    -(10d + Math.max(Math.abs(dx), Math.abs(dy)));
				} else {
					this._valueUpper = VALUE_MAX;	
				}
			}
			_priority = this._valueUpper - this._valueLower;
		}

		public void setMark() {
			_nMark = _nCurTime;
			//System.out.println("Mark " + this + " = " + _nMark);
		}
		
		public void unsetMark() {
			_nMark = _nCurTime - 1;
			//System.out.println("Mark " + this + " = " + _nMark);
		}
		
		public boolean isMarked() {
			//System.out.println("Is marked? " + this._nMark + " = " + _nCurTime);
			return _nMark == _nCurTime;
		}
		
		public State(State o) {
			this(o.x, o.y, o.dx, o.dy);
		}

		public String toString() {
			return "< x:" + x + ", y:" + y + ", dx:" + dx + ", dy:" + dy
					+ ", type: '" + ctype + "' [" + 
					_df.format(_valueLower) + ", " + 
					_df.format(_valueUpper) + "], VPI: " + _df.format(_maxVPI) + " >";
		}

		public void set(int x, int y, int dx, int dy) {
			this.x = x;
			this.y = y;
			this.dx = dx;
			this.dy = dy;
		}

		public double getReward(int act) {
			return -1d; // -1 cost for all Racetrack actions in any state 
		}
		
		// Returns best action during Bellman update
		// Note: if call multiple times, get different results due to looping states?
		//       thus can get different actions even when not updating second time
		public BestAction doBellmanUpdate(boolean do_update) {
			double max_value = Double.NEGATIVE_INFINITY;
			double max_value_lower = Double.NEGATIVE_INFINITY;
			double max_value_avg   = Double.NEGATIVE_INFINITY;
			double max_priority = 0d;
			
			// Compute expectation
			if (!hasNeighbors())
				populateNeighbors();

			int max_act = -1;
			int avg_act = -1;
			int min_act = -1;
			
			// System.out.print("Cur: " + _valueUpper);
			for (int act = 0; act < 9; act++) {
				
				if (_nextState[act] == null) {
					System.out.println("Skipping: " + act);
					continue;
				}
				
				double accum = getReward(act);
				double accum_lower = getReward(act);
				for (int next = 0; next < _nextState[act].length; next++) {
					State s_prime = _nextState[act][next];
					Solver.GOAL_FOUND = Solver.GOAL_FOUND || (s_prime.ctype == 'f'); // {System.out.println("\nFFF"); System.exit(1); }
					double prob = _nextProb[act][next];
					accum += prob * getMaxValue(s_prime);
					accum_lower += prob * getMinValue(s_prime);
				}
				
				// Following is only done for VPI-RTDP
				if (MAINTAIN_QVALUES) {
					_qLow[act] = accum_lower;
					_qUp[act]  = accum;
					if (!USE_UPPER_BOUND_COMP)
						_qAvg[act] = (accum + accum_lower)/2d;
				}
				
				// Note: Reason why accum_lower < VALUE_MIN...
				//       is due to the fact that MDP is non-discounted...
				//       no stationary max value... can always incur
				//       additional cost over current value
				//if (accum_lower < VALUE_MIN) {
					//System.out.println("Was: " + accum_lower + " changing to " + VALUE_MIN);
				//	accum_lower = VALUE_MIN;
				//}
				//if (accum < VALUE_MIN) {
				//	accum = VALUE_MIN;
				//}

				// Is best yet
				if (accum > max_value) {
					max_value = accum;
					max_act = act;
				}
				if (accum_lower > max_value_lower) {
					//System.out.println("Lower[act:" + act + "] = " + accum_lower);
					max_value_lower = accum_lower;
					min_act = act;
				}
				
				double avg = (accum_lower + accum)/2d;
				if (avg > max_value_avg) {
					max_value_avg = avg;
					avg_act = act;
				}
			}

			if (Double.isInfinite(max_value) || max_act == -1) {
				System.out.println("WTF" + max_value + ", " + 
						max_value_lower + ", " + max_act);
				System.exit(1);
			}

			// Make update
			double old_low = _valueLower;
			double old_high = _valueUpper;
			if (do_update) {
				_nBellmanUpdates++;
				_valueUpper = _valueUpper < max_value ? _valueUpper : max_value;
				_valueLower = _valueLower > max_value_lower ? _valueLower : max_value_lower;
			}

			if (_valueUpper < _valueLower /*|| Solver.GOAL_FOUND*/) {
				System.out.println("Prev values: [" + old_low + ", " + old_high + "]");
				System.out.println((Solver.GOAL_FOUND ? "Found finish: ": "Value mismatch: ") + this);
				System.exit(1);
			}
			
			if (_ps != null && _compare != null) {
				double actual_val = ((State)_compare._racetrack._states.get(this))._valueLower;
				double percent_range = (actual_val - _valueLower) / (_valueUpper - _valueLower);
				_ps.println(percent_range);
			}
			
			// Update priorities for FRTDP... requires max_act is true Q-maximizer
			if (MAINTAIN_FRTDP_PRIORITY) {
				for (int next = 0; next < _nextState[max_act].length; next++) {
					State s_prime = _nextState[max_act][next];
					double prob = _nextProb[max_act][next];
					double priority = prob * s_prime._priority;
					if (priority > max_priority)
						max_priority = priority;
				}
				_priority = Math.min(_priority, max_priority);
			}
						
			return new BestAction(min_act, avg_act, max_act);
		}

		public State chooseNextStateBoundGap(int act) {
			return chooseNextStateBoundGap(act, false, true);
		}
		
		public State chooseNextStateBoundGap(int act, boolean lookahead, boolean random) {

			if (_nextState[act] == null) {
				System.out.println("chooseNextStateVPI: chosen action " + act + " is null");
				System.exit(1);
			}
			
			// Draw a uniform point between 0 and 1
			double point = _random.nextDouble();
			double normal = 0.0;
			double best_priority = -1d;
			State best_s_prime = null; 
				
			/* compute the normalizing factor */
			for (int next = 0; next < _nextState[act].length; next++) {
				State s_prime = _nextState[act][next];
				if (SKIP_MARKED && (s_prime == null || s_prime.isMarked())) continue;
				if (s_prime == null && _nextProb[act][next] != 0.0) {
					System.out.println("Racetrack.chooseNextStateBoundGap: Non-zero prob of null state");
					System.exit(1);
				}
				double prob_trans = _nextProb[act][next];
				double priority = prob_trans * 
					(lookahead ? getLookaheadPriority(s_prime) : getBoundGap(s_prime));
				
				normal += priority;
				
				if (priority > best_priority) {
					best_priority = priority;
					best_s_prime = s_prime;
				}
			}
			
			if (normal < BRTDP_TOLERANCE) {
				return null;
			}
			
			if (!random) return best_s_prime;
			
			// Find s_prime interval where point falls
			for (int next = 0; next < _nextState[act].length; next++) {
				State s_prime = _nextState[act][next];
				if (SKIP_MARKED && (s_prime == null || s_prime.isMarked())) continue;
				//if (s_prime == null || s_prime.isMarked()) continue;
				double prob_choose = _nextProb[act][next] * 
				(lookahead ? getLookaheadPriority(s_prime) : getBoundGap(s_prime)) / normal;
				if (point <= prob_choose) {
					// Return state
					return s_prime;
				}

				// Skip to next interval
				point -= prob_choose;
			}

			System.out.println("chooseNextStateBiased: Should not get here... "
					+ "probabilities do not sum to 1!!!");
			System.exit(1);
			return null;
		}
		
		public State chooseNextStateVPI(BestAction act) {
					
			if (_nextState[act.ub_action] == null) {
				System.out.println("chooseNextStateVPI: chosen action " + act + " is null");
				System.exit(1);
			}
				
			// Note: this can be a lot more efficient if the correct values
			//       are cached
			//System.out.println("\n--------------------------------------------------" + 
			//		"\nStarting in: " + toString() + 
			//		"\n--------------------------------------------------");

			double max_vpi = -1d;
			State max_vpi_state = null;
			double sum_vpi = 0d;

			for (int next = 0; next < _nextState[act.ub_action].length; next++) {
				
				State t = _nextState[act.ub_action][next];
				
				// Check for probability of zero or mark
				if ((_nextProb[act.ub_action][next] == 0.0) || (SKIP_MARKED && t.isMarked())) {
					continue;
				}
	
				// Early on, waste VPI computation when bounds are large
				// This could be made more efficient by caching max bound gap
				// state and gap value during the Bellman backup...
				if (t._valueUpper - t._valueLower > BOUND_GAP_CUTOFF)
					return chooseNextStateBoundGap(act.ub_action);
				
				t._maxVPI = 0d;

				// *Note*: Empirically using Q-AVG and using the Q-UB value yield
				//         the same performance (actually a little better for UB
				//         since can save space with less memory / computation
				//         overhead).  But the formal theory calls for using AVG,
				//         so that is currently implemented here.
				
				// Compute c_{t,a*} and d_{t,a*}
				// d_{t,a*} is just the upper bound action probability
				//          value varies between lower and upper bound
				// c_{t,a*} is just the AVG Q-value for the UB action - (d_{t,a*} * t.AVG_VAL)
				double d_ta_star = _nextProb[act.ub_action][next];
				double c_ta_star = getReward(act.ub_action);
				if (USE_UPPER_BOUND_COMP)
					c_ta_star += this._qUp[act.ub_action] - (d_ta_star * (t._valueUpper + t._valueLower)/2d);
				else
					c_ta_star += this._qAvg[act.ub_action] - (d_ta_star * (t._valueUpper + t._valueLower)/2d);
				
				// Iterate through all actions != a*
				//System.out.println("\n-----------");
				for (int a = 0; a < 9; a++) {
					if (a == act.ub_action) continue;
					
					// Compute c_{t,a} and d_{t,a}					
					double d_ta = 0d;
					double c_ta = getReward(a);
					for (int next2 = 0; next2 < _nextState[a].length; next2++) {
						if (t.equals(_nextState[a][next2])) 
							{ d_ta = _nextProb[a][next2]; break; }
						
						// Old version of code... noticeable performance improvement
						// is use cached Q-values instead
						//
						//if (!t.equals(_nextState[a][next2])) 
						//	c_ta += _nextProb[a][next2] * _nextState[a][next2]._valueUpper;
						//else 
						//	d_ta = _nextProb[a][next2];
					}
					if (USE_UPPER_BOUND_COMP)
						c_ta += this._qUp[a] - (d_ta * (t._valueUpper + t._valueLower)/2d);
					else
						c_ta += this._qAvg[a] - (d_ta * (t._valueUpper + t._valueLower)/2d);
					
					// Following can occur b/c a state can reach itself and then it's
					// value was updated... ignoring this possibility here, empirically
					// has no impact on performance.
					//
					// if (this._qUp[a] != c_ta + d_ta * t._valueUpper) {
					//	System.out.println("Values don't add up: " + this._qUp[a] + " =? " + (c_ta + d_ta * t._valueUpper));
					//	System.exit(1);
					// }
					
					// This is a more general computation... can specialize for UB action
					// Compute pairwise VPI
					double vpi = computeVPI(c_ta_star, d_ta_star, c_ta, d_ta, 
							t._valueLower, t._valueUpper);
					
					/* This can be used if using UB Q-Values, but not AVG Q-Values
					double vpi = 0d; //(t._valueUpper - t._valueLower);
					
					// Missing a 1/2, but this is all relative
					if (Math.abs(d_ta_star - d_ta) > 1e-3d) {
						double x_pt = (c_ta - c_ta_star) / (d_ta_star - d_ta);
						double q_lta_minus_q_lta_star = c_ta - c_ta_star + t._valueLower * (d_ta - d_ta_star);
						if (q_lta_minus_q_lta_star > 0d)
							vpi = q_lta_minus_q_lta_star * (x_pt - t._valueLower) / (t._valueUpper - t._valueLower);
					}
					*/
					
					if (vpi > t._maxVPI) {
						t._maxVPI = vpi;
					}
					
					if (c_ta_star < 2d*VALUE_MIN || d_ta_star < 0 ||
							c_ta < 2d*VALUE_MIN || d_ta < 0 || t._maxVPI < 0) {
						System.out.println(c_ta_star + ", " +  d_ta_star + ", " + 
							c_ta + ", " +  d_ta + ", " + t._maxVPI);
						System.out.println("Error in chooseNextStateVPI");
						System.exit(1);
					}

					//System.out.println("  (" + c_ta_star + ", " +  d_ta_star + "), (" + 
					//		c_ta + ", " +  d_ta + "), [" + getMinValue(t) + ", " + getMaxValue(t) + "], --> VPI: " + 
					//		all_vpi[index]);
					//System.out.println("  a_star = " + getActionString(act) + ", a = " + getActionString(a));
				}
				
				//System.out.println("VPI[" + next + "] = " + t._maxVPI + " : " + t);
				sum_vpi += t._maxVPI;
				//System.out.println("Adding: " + all_vpi[index] + " -> " + sum_vpi);
				//System.out.println("sum += [" + index + "] = " + all_vpi[index]);
				if (t._maxVPI > max_vpi) {
					max_vpi = t._maxVPI;
					max_vpi_state = t;
				}
			}
						
			// For convergence, need probability of continuation
			if (sum_vpi <= VPI_CUTOFF) {
				if (_random.nextDouble() < ALPHA)
					return chooseNextStateBoundGap(act.ub_action);
				else
					return null;
			}
			
			if (!DRAW_RANDOM_VPI)
				return max_vpi_state;
			
			double point = _random.nextDouble();			
			for (int next = 0; next < _nextState[act.ub_action].length; next++) {
				
				State s_prime = _nextState[act.ub_action][next];
				
				// Check for probability of zero or mark
				if ((_nextProb[act.ub_action][next] == 0.0) || (SKIP_MARKED && s_prime.isMarked())) {
					continue;
				}

				double prob_choose = s_prime._maxVPI / sum_vpi;
				//System.out.println("Prob[" + next + "]: " + prob_choose + " = " + s_prime._maxVPI + " / " + sum_vpi + ": " + s_prime);
				
				if (point <= prob_choose) {
					// Return state
					//System.exit(1);
					return s_prime;
				}

				// Skip to next interval
				point -= prob_choose;
			}
			
			System.out.println("chooseNextStateVPI: Should not get here... "
					+ "probabilities do not sum to 1!!!");
			System.exit(1);
			return null;
		}

		double computeVPI(double c_ta_star, double d_ta_star, 
						  double c_ta, double d_ta, 
						  double lower, double upper) {
		
			if (Math.abs(d_ta_star - d_ta) < 1e-3d) 
				return 0d;

			double x_pt = (c_ta - c_ta_star) / (d_ta_star - d_ta);
			//System.out.println("x_pt = " + x_pt);
			//System.out.println("       [" + (c_ta_star + lower * d_ta_star) + ", " + (c_ta_star + upper * d_ta_star) + "]");
			//System.out.println("       [" + (c_ta + lower * d_ta) + ", " + (c_ta + upper * d_ta) + "]");

			double q_lta_minus_q_lta_star = c_ta - c_ta_star + lower * (d_ta - d_ta_star);
			double q_hta_minus_q_hta_star = c_ta - c_ta_star + upper * (d_ta - d_ta_star);

			if (x_pt <= lower || x_pt >= upper) {
				
				// One action could still completely dominate the other
				// This is more general than needed for the UB case... see code below
				
				// Test a point in middle
				double q_mta_minus_q_mta_star = c_ta - c_ta_star + ((lower + upper)/2d) * (d_ta - d_ta_star);
				if (q_mta_minus_q_mta_star > 0d) {
					// a dominates a_star*
					// area is h * (a+b) / 2 ... a,b are parallel sides
					//System.out.println("  Dominance: " + q_lta_minus_q_lta_star + ", " + q_hta_minus_q_hta_star + ", " + ((q_lta_minus_q_lta_star + q_hta_minus_q_hta_star)/2d));
					return (q_lta_minus_q_lta_star + q_hta_minus_q_hta_star)/2d;
					// * (upper - lower) / (upper - lower); height and uniform normalizer cancel
				} else // dominated
					return 0d;
			}
			
			// Bias distribution according to 1/8 interval histogram?

			/*
			double mid_point_mult = 1d;
			if (false) {
				double range_3_4 = 3d * (upper - lower)/4d;
				double range_7_8 = range_3_4 + (upper - lower)/8d;
				double range_15_16 = range_7_8 + (upper - lower)/16d;
				if ((x_pt >= lower + range_3_4 && q_hta_minus_q_hta_star > 0d)) {
					mid_point_mult = 2d;				
				}
				if ((x_pt >= lower + range_7_8 && q_hta_minus_q_hta_star > 0d)) {
					mid_point_mult = 4d;				
				}
				if ((x_pt >= lower + range_15_16 && q_hta_minus_q_hta_star > 0d)) {
					mid_point_mult = 8d;				
				}
			}
			*/

			if (q_lta_minus_q_lta_star > 0d)
				return q_lta_minus_q_lta_star * (x_pt - lower) / (upper - lower);
			else
				return q_hta_minus_q_hta_star * (upper - x_pt) / (upper - lower);
		}

		// Special case when astar is known to be UB action
		double computeVPI_UB_AStar(double c_ta_star, double d_ta_star, 
						  double c_ta, double d_ta, 
						  double lower, double upper) {
		
			double vpi = 0d;
			
			// Missing a 1/2, but this is all relative
			if (Math.abs(d_ta_star - d_ta) > 1e-3d) {
				double x_pt = (c_ta - c_ta_star) / (d_ta_star - d_ta);
				double q_lta_minus_q_lta_star = c_ta - c_ta_star + lower * (d_ta - d_ta_star);
				if (q_lta_minus_q_lta_star > 0d)
					vpi = q_lta_minus_q_lta_star * (x_pt - lower) / (upper - lower);
			}
			
			return vpi;
		}
		
		public double getBoundGap(State t) {
			return getMaxValue(t) - getMinValue(t);
		}
		
		// Should use cached priorities
		public double getLookaheadPriority(State t) {
			if (!MAINTAIN_FRTDP_PRIORITY) {
				System.out.println("Not maintaining priority... cannot retrieve");
				System.exit(1);
			}
			return t._priority;
		}

		public double getMaxValue(State t) {
			if (t == null) return VALUE_MAX;
			return t._valueUpper;
		}

		public double getMinValue(State t) {
			if (t == null) return VALUE_MIN;
			return t._valueLower;
		}

		public double getAvgValue(State t) {
			return (getMaxValue(t) + getMinValue(t)) / 2d;
		}
		
		public State chooseNextState(int act) {
			// Draw a uniform point between 0 and 1
			double point = _random.nextDouble();
			double normal = 0.0;
			// Find s_prime interval where point falls
			for (int next = 0; next < _nextState[act].length; next++) {
				State s_prime = _nextState[act][next];
				double prob_trans = _nextProb[act][next];

				if (point <= prob_trans) {
					// Return state
					//System.out.println("Chose next state: " + next);
					return s_prime;
				}

				// Skip to next interval
				point -= prob_trans;
			}

			System.out.println("chooseNextState: Should not get here... "
					+ "probabilities do not sum to 1!!!");
			System.exit(1);
			return null;
		}

		public boolean equals(Object other) {
			if (other == this)
				return true;
			if (other == null)
				return false;
			if (getClass() != other.getClass())
				return false;
			State s = (State) other;
			return (x == s.x && y == s.y && dx == s.dx && dy == s.dy);
		}

		public int hashCode() {
			return x + (y << 10) + (dx << 18) + (dy << 24);
		}

		public int compareTo(Object other) {
			if (other == this)
				return 0;
			if (other == null)
				return 1;
			if (getClass() != other.getClass())
				return this.toString().compareTo(other.toString());
			State s = (State) other;
			if (x != s.x) {
				return x - s.x;
			} else if (y != s.y) {
				return y - s.y;
			} else if (dx != s.dx) {
				return dx - s.dx;
			} else if (dy != s.dy) {
				return dy - s.dy;
			} else 
				return 0;
		}

		public boolean hasNeighbors() {
			return _nextState != null;
		}
		
		// Helper object and function for populate Neighbors
		HashMap<NextState,NextState> succ_map = new HashMap<NextState,NextState>();
		public void addToSuccMap(State n, double prob) {
			NextState s = new NextState(prob, n);
			NextState curr = null;
			if ((curr = succ_map.get(s)) == null) {
				succ_map.put(s, s);
			} else {
				curr.add(s);
			}
		}
		
		public void populateNeighbors() {
			if (_nextState != null)
				return;

			_nextState = new State[9][];
			_nextProb = new double[9][];
			
			// Following is only done for VPI-RTDP
			if (MAINTAIN_QVALUES) {
				_qLow = new double[9];
				_qUp  = new double[9];
				if (!USE_UPPER_BOUND_COMP)
					_qAvg = new double[9];
			}
				
			for (int act = 0; act < 9; act++) {
				int ddx = (act % 3) - 1;
				int ddy = ((act / 3) % 3) - 1;

				int dx_temp = dx + ddx;
				int dy_temp = dy + ddy;
				_tempState.set(x + dx + ddx, y + dy + ddy, dx + ddx, dy + ddy);
				char temp_ctype = _track.get(_tempState.x, _tempState.y);

				/* NOISY RACETRACK TRANSITIONS */

				succ_map.clear();			 
				addToSuccMap(this, 0.10); 
				
				 _tempState.ctype = temp_ctype;
				 addToSuccMap(getState(_tempState, this), 0.10); 
				 _tempState.dx+=1;
				 addToSuccMap(getState(_tempState, this), 0.20);
				 _tempState.dx-=2; 
				 addToSuccMap(getState(_tempState, this), 0.20); 
				 _tempState.dy+=1;
				 _tempState.dx+=1;
				 addToSuccMap(getState(_tempState, this), 0.20);
				 _tempState.dy-=2; 
				 addToSuccMap(getState(_tempState, this), 0.20); 
				 
				 _nextState[act] = new State[succ_map.size()]; 
				 _nextProb[act] = new double[succ_map.size()]; 
				 int index = 0;
				 for (NextState s : succ_map.keySet()) {
					 _nextState[act][index] = s.state;
					 _nextProb[act][index++] = s.prob;
				 }
				 
				 // Normalize probabilities
				 double sum_prob = 0.0;
				 for (int j = 0; j < _nextState[act].length; j++)
					 sum_prob += _nextProb[act][j];
				 if (sum_prob == 0.0)
					 _nextState[act] = null;
				 else 
					 for (int j = 0; j < _nextState[act].length; j++) {
						 _nextProb[act][j] /= sum_prob;
						 //System.out.println( _nextProb[act][j] + ": " + _nextState[act][j]);
					 }
			}
		}

		// Return a legal cached state for _tempState
		State getState(State _tempState, State def_state) {
			State ret = getState(_tempState);
			if (ret != null) return ret; //else return def_state;
			
			// State was out of bounds, so just stop in current state
			State s = new State(def_state);
			s.dx = 0;
			s.dy = 0;
			return getState(s);
		}
			
		State getState(State _tempState) {
			if (_track.get(_tempState.x, _tempState.y) == '@')
				return null;
			
			State temp = (State) _states.get(_tempState);
			if (temp == null) {
				temp = new State(_tempState);
				_states.put(temp, temp); // Store in cache for reuse
			} // otherwise already in cache
			return temp;
		}

	}

	////////////////////////////////////////////////////////////////////////////
	// Racetrack Helper Functions
	////////////////////////////////////////////////////////////////////////////

	public Racetrack(String filename) {
		_filename = filename;
		readTrack(filename);
	}

	public void addStateAndPopulateNeighbors(int x, int y, int dx, int dy) {
		_tempState.set(x, y, dx, dy);
		State temp = (State) _states.get(_tempState);
		if (temp == null) {
			temp = new State(_tempState);
			temp.ctype = _track.get(_tempState.x, _tempState.y); // Set surface
			// type char
			_states.put(temp, temp); // Store in cache for reuse
		}
		if (!temp.hasNeighbors())
			temp.populateNeighbors();
	}

	public double computeInitStateLAvg() {
		double sum = 0.0;
		Iterator i = _startStates.iterator();
		while (i.hasNext()) {
			State s = (State) i.next();
			sum += s._valueLower;
		}
		return sum / _startStates.size();
	}

	public double computeInitStateLUDiff() {
		double error = 0.0;
		Iterator i = _startStates.iterator();
		while (i.hasNext()) {
			State s = (State) i.next();
			error += s._valueUpper - s._valueLower;
		}
		return error / _startStates.size();
	}

	public double computeInitStateError() {
		double error = 0.0;
		Iterator i = _startStates.iterator();
		while (i.hasNext()) {
			State s = (State) i.next();
			error += Math.abs(s._valueUpper - s._valueLower);
		}
		return error / _startStates.size();
	}

	public State chooseInitialState() {
		// System.out.println(_startStates);
		// System.exit(1);
		int num_start_states = _startStates.size();
		int random_start_index = _random.nextInt(num_start_states);
		return (Racetrack.State) _startStates.get(random_start_index);
	}

	public void readTrack(String filename) {

		try {
			ArrayList al = new ArrayList();
			BufferedReader reader = new BufferedReader(new FileReader(filename));
			String line = null;
			String last_line = "";
			boolean reach_dash = false;
			while ((line = reader.readLine()) != null) {
				reach_dash |= (last_line.trim().equals("---"));
				if (reach_dash) {
					al.add(line);
					System.out.println("'" + line + "'");
				}
				last_line = line;
			}

			_track = new Track(al.size(), ((String) al.get(0)).length());
			for (int i = 0; i < al.size(); i++) {
				_track._layout[i] = (String) al.get(i);
			}

			collectStartStates();
			setFinalState();
			if (filename.indexOf("bl") > 0)
				BOUND_GAP_CUTOFF = VALUE_MAX - VALUE_MIN; 
				

		} catch (Exception e) {
			System.out.println("readTrack: " + e);
			System.exit(1);
		}
	}

	public static class Pos {
		public Pos(int px, int py) { this.x = px; this.y = py; }
		int x;
		int y;
	}
	
	public void setFinalState() {
		for (int l = 0; l < _track._height; l++) {
			for (int j = 0; j < _track._width; j++) {
				char c = _track._layout[l].charAt(j);
				if (c == 'f') {
					_finalPos = new Pos(j, l);
					return;
				}
			}
		}
	}

	public void collectStartStates() {
		_startStates.clear();
		for (int l = 0; l < _track._height; l++) {
			for (int j = 0; j < _track._width; j++) {
				char c = _track._layout[l].charAt(j);
				if (c == 's') {
					_tempState.set(j, l, 0, 0); // Start with 0 velocity
					State temp = new State(_tempState);
					// System.out.println("Start: " + j + ", " + l + ": " +
					// _tempState + ", " + temp);
					temp.ctype = 's';
					_states.put(temp, temp);
					_startStates.add(temp);
				}
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////
	// Graphics Code Below
	////////////////////////////////////////////////////////////////////////////

	public void displayInitial() {

		if (_track == null)
			return;

		// Initialize the frame
		bd = new BlockDisplay("Racetrack: " + _filename, "", _track._height,
				_track._width);
		System.out.println("height = " + _track._height + ", width = "
				+ _track._width);

		// Draw a block of cells
		bd.clearAllCells();
		bd.clearAllLines();
		for (int l = 0; l < _track._height; l++) {
			for (int j = 0; j < _track._width; j++) {
				char c = _track._layout[l].charAt(j);
				// System.out.println(c + " at " + "(" + l + "," + j + ")");
				switch (c) {
				case '@':
					bd.setCell(l, j, Color.black, null);
					break;
				case 's':
					bd.setCell(l, j, Color.green, null);
					break;
				case 'f':
					bd.setCell(l, j, Color.red, null);
					break;
				case ' ':
					bd.setCell(l, j, Color.white, null);
					break;
				case 'x':
					bd.setCell(l, j, Color.white, null);
					break;
				// case 'x': bd.setCell(l,j,Color.gray, "x"); break;
				default:
					bd.setCell(l, j, Color.yellow, null);
					break;
				}
			}
		}

		// bd.addLine(Color.blue, 6, 6, 20, 20);
		// bd.addLine(Color.red, 10, 10, 10, 20);
		// bd.addLine(Color.red, 10, 10, 20, 10);
		// bd.setMessage("Message");
		bd.repaint();

	}

	public void restoreBackground(State s) {
		switch (s.ctype) {
		case '@':
			bd.setCell(s.y, s.x, Color.black, null);
			break;
		case 's':
			bd.setCell(s.y, s.x, Color.green, null);
			break;
		case 'f':
			bd.setCell(s.y, s.x, Color.red, null);
			break;
		case ' ':
		case 'x':
			bd.setCell(s.y, s.x, Color.white, null);
			break;
		// case 'x': bd.setCell(s.y,s.x,Color.gray, "x"); break;
		default:
			bd.setCell(s.y, s.x, Color.yellow, null);
			break;
		}
	}

	public double runPolicy(String msg) {

		// Select start state at random
		Racetrack.State s = chooseInitialState();
		
		if (Solver.VISUALIZE_POLICY) {
			bd.setMessage(msg);
			bd.setCell(s.y, s.x, Color.cyan, null);
			bd.repaint();
		}

		// Execute trajectory until max length reached or success
		// Note: assume no discounting (we have bounded total reward)
		double accum_reward = 0.0;
		for (int actions = 0; actions < Racetrack.MAX_TRAJ_LENGTH
				&& s.ctype != 'f'; actions++) {

			// Check if neighbors of s populated
			if (!s.hasNeighbors()) {
				s.populateNeighbors();
			}

			// Get greedy action for upper bound
			BestAction act = s.doBellmanUpdate(false /* do not do update */);
			// Take action and randomly pick next state
			State old = s;
			s = s.chooseNextState(act.lb_action);
			if (!s.equals(old) && Solver.VISUALIZE_POLICY) {
				restoreBackground(old);
				bd.setCell(s.y, s.x, Color.cyan, null);
				bd.repaint();

				try {
					Thread.currentThread().sleep(5);
				} catch (InterruptedException e) {
					System.out.println(e);
				}
			}

			// Decrement reward for action taken
			accum_reward = accum_reward - 1.0;
		}

		if (Solver.VISUALIZE_POLICY) {
			restoreBackground(s);
			bd.repaint();
		}
		return accum_reward;
	}

	public void killDisplay() {

		// Kill window
		bd.close();
	}

	public static void main(String[] args) {

		Racetrack r = new Racetrack("small-b-p2.racetrack");
		// Racetrack r = new Racetrack("large-b-3.racetrack");
		// Racetrack r = new Racetrack("hall4.racetrack");
		// Racetrack r = new Racetrack("large-ring4.racetrack");
		// Racetrack r = new Racetrack("hall5-w2.racetrack");

		// Sleep
		try {
			r.displayInitial();
			Thread.currentThread().sleep(5000);
			r.killDisplay();
		} catch (InterruptedException e) {
			System.out.println(e);
			System.exit(1);
		}

	}

}
