package mdp.rtdp;

import java.util.*;
import java.io.*;
import java.text.*;

import util.Timer;

public abstract class Solver {

	public static boolean VISUALIZE_SEARCH = false;
	public static boolean VISUALIZE_POLICY = false;
	public static boolean GOAL_FOUND = false;
	
	public static int POLICY_TRIALS = 300; // Variance of this is very stable
	public static int NUM_RUNS = 30; // 60, 1
	public static int NUM_TRIALS_PER_RUN = 4000; // Not used if NUM_MS_PER_RUN used
	public static int DISPLAY_INTERVAL = 50; // 50, 250
	public static int NUM_MS_PER_RUN = 10000;
	public static final int MIN_DISPLAY_TRIALS = 1;
	public static final boolean RECORD_VALUES = false;
	
	public static DecimalFormat _df = new DecimalFormat("#.##");
	
	public int _nCurRun = -1;
	public int _nCurTrial = -1;	
	public Racetrack _racetrack;
	public Timer _timer = new Timer();

	public Solver(Racetrack r) {
		_racetrack = r;
		_racetrack._states.clear(); // Clear all states and values
	}
	
	public abstract String getName();
	
	public ArrayList[] runTrials(int num_trials, boolean use_time_limit) {
		
		// Reset problem
		_racetrack._states.clear();
		_racetrack.collectStartStates();

		// Run num_trials trials
		Racetrack._nBellmanUpdates = 0;
		double accum_reward = 0.0;
		double accum_time = 0.0;
		ArrayList trial_num_list = new ArrayList();
		ArrayList reward_list = new ArrayList();
		ArrayList std_list = new ArrayList();
		ArrayList numstates_list = new ArrayList();
		ArrayList avg_init_state_error_list = new ArrayList();
		ArrayList bellman_updates = new ArrayList();
		ArrayList avg_time_list = new ArrayList();

		// Garbage collect before running trials to avoid delayed memory
		// effects of previous runs
		Runtime.getRuntime().gc();
		
		for (_nCurTrial = 1; use_time_limit ? accum_time < 1.5d * num_trials : _nCurTrial <= num_trials; _nCurTrial++) {
			
			// Run an individual training trial

			_timer.ResetTimer();
			double reward = runTrial();
			accum_time += _timer.StopTimer();
			//System.out.println("Trial " + _nCurTrial + " finished: " + reward);
			accum_reward += reward;
			
			// Incremental output and policy simulation
			if (_nCurTrial % DISPLAY_INTERVAL == 0 && _nCurTrial >= MIN_DISPLAY_TRIALS) {
				
				// Terminal output
				//System.out.println("****************************************");
				double avg_reward = accum_reward / DISPLAY_INTERVAL;
				trial_num_list.add(new Double(_nCurTrial));
				//reward_list.add(new Double(avg_reward));
				numstates_list.add(new Double(_racetrack._states.size()));
				bellman_updates.add(new Double(_racetrack._nBellmanUpdates));
				double avg_init_state_error = _racetrack.computeInitStateError(); 
				avg_init_state_error_list.add(new Double(avg_init_state_error));
				avg_time_list.add(accum_time /* / _nCurTrial*/);
				accum_reward = 0.0;
				//accum_time = 0.0;
				//System.out.println("****************************************");
				
				// Simulate policy
				double pol_accum = 0.0;
				ArrayList pol_rewards = new ArrayList();
				for (int j = 0; j < POLICY_TRIALS; j++) {
					double pol_reward = _racetrack.runPolicy("Policy as of Trial #" + _nCurTrial + 
							" (#states = " + _racetrack._states.size() +  
							", trial avg to term. = " + avg_reward + 
							", init error = " + _df.format(avg_init_state_error) + ")");
					pol_accum += pol_reward;
					pol_rewards.add(new Double(pol_reward));
				}
				double pol_avg = pol_accum / POLICY_TRIALS;
				reward_list.add(new Double(pol_avg));
				double pol_std = computeStd(pol_rewards);
				std_list.add(new Double(pol_std));
				
				System.out.println(getName() + " [#" + _nCurRun + "/" + 
						NUM_RUNS + ", " + _nCurTrial + 
						", " + _df.format(accum_time) + "ms]: avg policy reward = " + 
						_df.format(pol_avg) + " ~ " + _df.format(pol_std));
			}
		}
		
		return new ArrayList[] {trial_num_list, reward_list, std_list, avg_init_state_error_list,
				numstates_list, bellman_updates, avg_time_list} ;
	}
	
	public abstract double runTrial();
	
	public static double computeMean(Object val) {
		if (val instanceof Double)
			return ((Double) val).doubleValue();
		ArrayList values = (ArrayList)val;
		
		double mean = 0d, var = 0d;
		for (int i = 0; i < values.size(); i++)
			mean += (Double)values.get(i);
		mean /= values.size();
		
		return mean;
	}

	public static double computeStd(Object val) {
		if (val instanceof Double)
			return 0d;
		ArrayList values = (ArrayList)val;
		
		double mean = 0d, var = 0d;
		for (int i = 0; i < values.size(); i++)
			mean += (Double)values.get(i);
		mean /= values.size();
		
		for (int i = 0; i < values.size(); i++)
			var += Math.pow(((Double)values.get(i) - mean), 2d);		
	
		var = Math.sqrt((var / (values.size() *(values.size() - 1))));
		
		if (Double.isInfinite(var) || Double.isNaN(var))
			var = 0d;
		
		return var;
	}
	 
	// n=30, 90%: 1.70
	// n=30, 95%: 2.04
	// n=30, 99%: 2.75
	// n=infty, 90%: 1.64
	// n=infty, 95%: 1.96
	// n=infty, 99%: 2.58
	public static final double CI = 1.96d;
	
    public static void printResults(PrintStream ps, ArrayList[] merge, String header) {
    	if (merge == null) return;
		if (header != null) {
			ps.println(header + "\n");
			ps.println("TRIAL\tERROR\t+-95%\tINI ERR\t#STATES\tBellUp\tAccTime");
			ps.println("==================================================");
		}
		for (int i = 0; i < merge[0].size(); i++) {
			ps.print(computeMean(merge[0].get(i)));
			ps.print("\t" + _df.format(computeMean(merge[1].get(i))));
			ps.print("\t" + _df.format(CI * computeStd(merge[1].get(i))));
			ps.print("\t" + _df.format(computeMean(merge[3].get(i))));
			ps.print("\t" + _df.format(computeMean(merge[4].get(i))));			
			ps.print("\t" + _df.format(computeMean(merge[5].get(i))));			
			ps.println("\t" + _df.format(computeMean(merge[6].get(i))));			
		}
    }
    
    public static void mergeResults(ArrayList[] l1, ArrayList[] l2) {
		for (int i = 0; i < Math.min(l1.length, l1.length); i++) {
			for (int j = 0; j < Math.min(l1[i].size(),l2[i].size()); j++) {
				Object o = l1[i].get(j);
				if (o instanceof Double) {
					ArrayList a = new ArrayList();
					a.add(o);
					o = a;
				}
				((ArrayList)o).add(l2[i].get(j));
				l1[i].set(j, o);
			}
		}
    }

    public static void printStates(Racetrack r) {    	
    	System.out.println("\nStates:");
    	for (Object o : r._states.values()) {
    		if (((Racetrack.State)o).dx == 0 && ((Racetrack.State)o).dy == 0)
    			System.out.println(o);
    	}
    }
    
	////////////////////////////////////////////////////////////////
	
    public static void main(String[] args) {

    	if (args.length < 8 || args.length > 10) {
    		System.out.println("\nOnly " + args.length + " cmd line options:");
    		for (int i = 0; i < args.length; i++) {
    			System.out.print(args[i] + " ");
    		}
    	    System.out.println("\n\nMust enter cmd line options in following order:");
    	    System.out.println("[0] RACETRACK_FILENAME={file ending in .racetrack}"); 
    	    System.out.println("[1] ALG={RTDP,BRTDP,FRTDP,VPIRTDP,VI,VIR}");
    	    System.out.println("[2] POLICY_TRIALS=int");
    	    System.out.println("[3] NUM_RUNS=int");
    	    System.out.println("[4] MS_PER_RUN=int"); 
    	    System.out.println("[5] TRIAL_DISP_INTERVAL=int"); 
    	    System.out.println("[6] SKIP_MARKED={true,false}"); 
    	    System.out.println("[7] INFORMED_HEURISTIC={true,false}"); 
    	    System.out.println("[8] VISUALIZE_SEARCH={true,DEF:false}"); 
    	    System.out.println("[9] VISUALIZE_POLICY={true,DEF:false}");
    	    System.exit(1);
    	}

    	// Parse options
    	try {
    		POLICY_TRIALS = Integer.parseInt(args[2]);
    		NUM_RUNS = Integer.parseInt(args[3]);
    		NUM_MS_PER_RUN = Integer.parseInt(args[4]);
    		DISPLAY_INTERVAL = Integer.parseInt(args[5]);
    		Racetrack.SKIP_MARKED = Boolean.parseBoolean(args[6]);
    		Racetrack.USE_INFORMED_HEURISTIC = Boolean.parseBoolean(args[7]);
    		if (args.length >= 9)  VISUALIZE_SEARCH = Boolean.parseBoolean(args[8]);
    		if (args.length >= 10) VISUALIZE_POLICY = Boolean.parseBoolean(args[9]);
    	} catch (Exception e) {
    	    System.exit(1);
    	}
    	  

    	// Parse racetrack filename
    	String racetrack_file = args[0]; // line, line2, line4, block, block2, block3
    	                                 // small-b, large-b, large-ring, hall-
    	Racetrack r = new Racetrack("racetrack/" + racetrack_file + ".racetrack");

    	Solver s = null;
    	if (args[1].equalsIgnoreCase("RTDP")) {
    		s = new RTDP(r);
    	} else if (args[1].equalsIgnoreCase("BRTDP")) {
    		s = new BRTDP(r);   
    	} else if (args[1].equalsIgnoreCase("FRTDP")) {
    		s = new FRTDP(r);
    	} else if (args[1].equalsIgnoreCase("VPIRTDP")) {
    		s = new VPIRTDP(r);
    	} else if (args[1].equalsIgnoreCase("VI")) {
    		s = new ValueIter(r);
    	} else if (args[1].equalsIgnoreCase("VIR")) {
    		s = new ValueIterReach(r);
    	} else {
    	    System.out.println("\nUnknown algorithm: " + args[1]);
    	    System.out.println("Required one of {RTDP,BRTDP,FRTDP,VPIRTDP,VI,VIR}");
    	    System.exit(1);
    	}
    	
    	if (VISUALIZE_SEARCH || VISUALIZE_POLICY)
    		r.displayInitial();

    	if (RECORD_VALUES) {
	    	Solver s2 = new ValueIter(new Racetrack(r._filename));
	    	s2.runTrials(100, false);
	    	s._racetrack._compare = s2;
			try {
				s._racetrack._ps = new PrintWriter(new FileOutputStream("mdp/rtdp/state_data.txt"));
			} catch (Exception e) {
				System.out.println(e);
				System.exit(1);
			}
    	}

    	ArrayList[] merge = null;
    	for (s._nCurRun = 0; s._nCurRun <= NUM_RUNS; s._nCurRun++) {
    		ArrayList[] results = s.runTrials(NUM_MS_PER_RUN, true);
        	//printResults(merge, "\nPre-merge 1");
        	//printResults(results, "\nPre-merge 2");
    		if (s._nCurRun == 0) continue;
    		if (merge == null)
    			merge = results;
    		else 
    			mergeResults(merge, results);
        	//printResults(merge, "\nPost-merge");
    	}
    	//normResults(merge, (double)NUM_RUNS);
    	//printStates(r);
    	printResults(System.out, merge, "\nAvg results");
    	
		try {
			PrintStream result_file = new PrintStream(
					new FileOutputStream(s.getName() + "_" + racetrack_file 
							+ (Racetrack.SKIP_MARKED ? "" : "_noskip") 
							+ (Racetrack.USE_INFORMED_HEURISTIC ? "" : "_uninform") + ".txt"));
			printResults(result_file, merge, null);
			result_file.close();
		} catch (Exception e) {
			System.out.println(e);
		}

    }
  
}
